import asyncio
import time
from multiprocessing import Process
from queue import Queue
from threading import Thread


def config_order_key(elm):
    return elm[1]['order_index']


class MessageQueue:

    def __init__(self, feed_processor) -> None:
        self.feed_processor = feed_processor
        self.worker_loop = asyncio.new_event_loop()
        self.__temp_queue__ = asyncio.queues.Queue()

    async def put(self, *args, **kwargs):
        await self.feed_processor(*args, *kwargs)
        # await self.__temp_queue__.put((args, kwargs))
        # await self.clear_process()

    async def clear_process(self):
        while not self.__temp_queue__.empty():
            temp_item_ = await self.__temp_queue__.get()
            await asyncio.ensure_future(self.feed_processor(*temp_item_[0], *temp_item_[1]))


class Reactor:
    __facts_senders__ = {}
    __facts_readers__ = {}
    __action_receivers__ = {}

    temp_buffer = {}
    latest_buffer = {}
    time_sync_source_name = None
    reactor_time = -1

    __message_queue__ = None

    def __init__(self) -> None:
        self.monitor = Reactor.Monitor()
        self.__message_queue__ = MessageQueue(self.__feed__processor__)

    class Monitor:
        index_flow_temp = []
        index_flow = ''
        index_changed = False

        pre_index = -1

        def index_monitor(self, index, source):
            if self.pre_index != index:
                self.index_flow = [str(s) for s in self.index_flow_temp]
                self.index_flow_temp = []
                self.index_changed = True
            else:
                self.index_changed = False

            self.index_flow_temp.append(source)

    class Context:

        def __init__(self, brain, fact_sender) -> None:
            self.fact_sender = fact_sender
            self.brain = brain

        async def get_current_value(self, source):
            return await self.brain.__read__(source=source)

        async def publish(self, index, data, source=None):
            source = source if source is not None else self.fact_sender
            await self.brain.__feed__(source=source, index=index, data=data)

    def fact_reader_register(self, name, fn, config):

        if name is not None and name in self.__facts_readers__:
            self.__facts_readers__[name].append((fn, config))
            self.__facts_readers__[name].sort(key=config_order_key)

    def fact_reader(self, _fn, name, source=None, sync_time=False, buffer_size=0, skip_size=1, run_separate=False,
                    sleep_time=0,
                    order_index=0):
        _fn.name = name
        _fn.sync_time = sync_time
        _fn.sleep_time = sleep_time
        _fn.order_index = order_index
        _fn.run_separate = run_separate

        if self.time_sync_source_name is None:
            self.time_sync_source_name = name

        self.__facts_senders__[name] = _fn

        if name not in self.__facts_readers__:
            self.__facts_readers__[name] = []

        if source is not None:
            config = {
                'buffer_size': buffer_size,
                'skip_size': skip_size,
                'order_index': order_index
            }
            self.fact_reader_register(source, _fn, config)

    def processor(self, name, source=None, sync_time=False, buffer_size=0, skip_size=1, run_separate=False,
                  sleep_time=0):
        def decorator(_fn):
            self.fact_reader(_fn=_fn, name=name, source=source, sync_time=sync_time, buffer_size=buffer_size,
                             skip_size=skip_size, run_separate=run_separate, sleep_time=sleep_time)

        return decorator

    async def __read__(self, source):
        if source in self.latest_buffer:
            return self.latest_buffer[source]
        return -1

    async def __feed__processor__(self, source, index, data):
        self.monitor.index_monitor(index, source)
        # print(index, source)
        self.latest_buffer[source] = (index, data)

        feed_source = self.__facts_senders__[source]

        if feed_source.sync_time:
            self.reactor_time = index

        for _reader_meta_ in self.__facts_readers__[source]:
            reader = _reader_meta_[0]
            reader_config = _reader_meta_[1]

            if reader_config['buffer_size'] > 0:
                buffer_size = reader_config['buffer_size']
                skip_size = reader_config['skip_size']
                temp_name = source + str(reader)
                if not temp_name in self.temp_buffer:
                    self.temp_buffer[temp_name] = Queue(maxsize=buffer_size)
                self.temp_buffer[temp_name].put({
                    'index': index,
                    'data': data
                })
                if self.temp_buffer[temp_name].full():
                    context = Reactor.Context(self, reader.name)
                    index = index
                    data_ = list(self.temp_buffer[temp_name].queue)
                    if not reader.run_separate:
                        await reader(context, index, data_)
                    else:
                        p = Thread(target=reader, args=(context, index, data_))
                        p.start()

                    for s in range(skip_size):
                        self.temp_buffer[temp_name].get()
            else:
                if not reader.run_separate:
                    context = Reactor.Context(self, reader.name)
                    await reader(context, index, data)
                else:

                    def start_function():
                        asyncio.run(reader(context, index, data))

                    p = Process(target=start_function)
                    p.start()
            time.sleep(feed_source.sleep_time)

    async def __feed__(self, source, index, data):

        # Sync with engine time
        if source is not self.time_sync_source_name and index > self.reactor_time:
            async def wait():
                while self.reactor_time < index:
                    # time.sleep(1e-50)
                    pass
                return True

            await wait()
        await self.__message_queue__.put(source, index, data)

    async def start(self):
        if self.time_sync_source_name is None:
            raise NotImplementedError("Time sync source cannot be empty")

        for fact_sender in self.__facts_senders__:
            sender_fn = self.__facts_senders__[fact_sender]
            context = Reactor.Context(self, fact_sender)
            if not sender_fn.run_separate:
                await sender_fn(context)
            else:
                def start_function():
                    asyncio.run(sender_fn(context))

                p = Thread(target=start_function)
                p.start()

    async def run(self):
        await self.start()



from pulathisi import Reactor


class Brain(Reactor):

    def process_sum(self, sources, _fn):
        pass

    def sum(self, name, sources):
        def decorator(_fn):
            @self.processor(name=name, source=self.time_sync_source_name)
            async def decorator_temp(context: Reactor.Context, index=-1, data=None):
                values = {}
                for source in sources:
                    source_data = await context.get_current_value(source=source)
                    values[source] = source_data
                await _fn(context, index, values)

        return decorator



from collections import deque
from typing import *

import asyncio
import collections
import itertools
import sys
import time
from typing import Any

import hiredis
import uvloop

expiration = collections.defaultdict(lambda: float("inf"))  # type: Dict[bytes, float]
dictionary = {}  # type: Dict[bytes, Any]


class RedisProtocol(asyncio.Protocol):
    def __init__(self):
        self.dictionary = dictionary
        self.response = collections.deque()
        self.parser = hiredis.Reader()
        self.transport = None  # type: asyncio.transports.Transport
        self.commands = {
            b"COMMAND": self.command,
            b"SET": self.set,
            b"GET": self.get,
            b"PING": self.ping,
            b"INCR": self.incr,
            b"LPUSH": self.lpush,
            b"RPUSH": self.rpush,
            b"LPOP": self.lpop,
            b"RPOP": self.rpop,
            b"SADD": self.sadd,
            b"HSET": self.hset,
            b"SPOP": self.spop,
            b"LRANGE": self.lrange,
            b"MSET": self.mset,
        }

    def connection_made(self, transport: asyncio.transports.Transport):
        self.transport = transport

    def data_received(self, data: bytes):
        self.parser.feed(data)

        while 1:
            req = self.parser.gets()
            if req is False:
                break
            else:
                self.response.append(self.commands[req[0].upper()](*req[1:]))

        self.transport.writelines(self.response)
        self.response.clear()

    def command(self):
        # Far from being a complete implementation of the `COMMAND` command of
        # Redis, yet sufficient for us to start using redis-cli.
        return b"+OK\r\n"

    def set(self, *args) -> bytes:
        # Defaults
        key = args[0]
        value = args[1]
        expires_at = None
        cond = b""

        largs = len(args)
        if largs == 3:
            # SET key value [NX|XX]
            cond = args[2]
        elif largs >= 4:
            # SET key value [EX seconds | PX milliseconds] [NX|XX]
            try:
                if args[2] == b"EX":
                    duration = int(args[3])
                elif args[2] == b"PX":
                    duration = int(args[3]) / 1000
                else:
                    return b"-ERR syntax error\r\n"
            except ValueError:
                return b"-value is not an integer or out of range\r\n"

            if duration <= 0:
                return b"-ERR invalid expire time in set\r\n"

            expires_at = time.monotonic() + duration

            if largs == 5:
                cond = args[4]

        if cond == b"":
            pass
        elif cond == b"NX":
            if key in self.dictionary:
                return b"$-1\r\n"
        elif cond == b"XX":
            if key not in self.dictionary:
                return b"$-1\r\n"
        else:
            return b"-ERR syntax error\r\n"

        if expires_at:
            expiration[key] = expires_at

        self.dictionary[key] = value
        return b"+OK\r\n"

    def get(self, key: bytes) -> bytes:
        if key not in self.dictionary:
            return b"$-1\r\n"

        if key in expiration and expiration[key] < time.monotonic():
            del self.dictionary[key]
            del expiration[key]
            return b"$-1\r\n"
        else:
            value = self.dictionary[key]
            return b"$%d\r\n%s\r\n" % (len(value), value)

    def ping(self, message=b"PONG"):
        return b"$%d\r\n%s\r\n" % (len(message), message)

    def incr(self, key):
        value = self.dictionary.get(key, 0)
        if type(value) is str:
            try:
                value = int(value)
            except ValueError:
                return b"-value is not an integer or out of range\r\n"
        value += 1
        self.dictionary[key] = str(value)
        return b":%d\r\n" % (value,)

    def lpush(self, key, *values):
        deque = self.dictionary.get(key, collections.deque())
        deque.extendleft(values)
        self.dictionary[key] = deque
        return b":%d\r\n" % (len(deque),)

    def rpush(self, key, *values):
        deque = self.dictionary.get(key, collections.deque())
        deque.extend(values)
        self.dictionary[key] = deque
        return b":%d\r\n" % (len(deque),)

    def lpop(self, key):
        try:
            deque = self.dictionary[key]  # type: collections.deque
        except KeyError:
            return b"$-1\r\n"
        value = deque.popleft()
        return b"$%d\r\n%s\r\n" % (len(value), value)

    def rpop(self, key):
        try:
            deque = self.dictionary[key]  # type: collections.deque
        except KeyError:
            return b"$-1\r\n"
        value = deque.pop()
        return b"$%d\r\n%s\r\n" % (len(value), value)

    def sadd(self, key, *members):
        set_ = self.dictionary.get(key, set())
        prev_size = len(set_)
        for member in members:
            set_.add(member)
        self.dictionary[key] = set_
        return b":%d\r\n" % (len(set_) - prev_size,)

    def hset(self, key, field, value):
        hash_ = self.dictionary.get(key, {})
        ret = int(field in hash_)
        hash_[field] = value
        self.dictionary[key] = hash_
        return b":%d\r\n" % (ret,)

    def spop(self, key):  # TODO add `count`
        try:
            set_ = self.dictionary[key]  # type: set
            elem = set_.pop()
        except KeyError:
            return b"$-1\r\n"
        return b"$%d\r\n%s\r\n" % (len(elem), elem)

    def lrange(self, key, start, stop):
        start = int(start)
        stop = int(stop)
        try:
            deque = self.dictionary[key]  # type: collections.deque
        except KeyError:
            return b"$-1\r\n"
        l = itertools.islice(deque, start, stop)
        return b"*%d\r\n%s" % (stop - start, b"".join(b"$%d\r\n%s\r\n" % (len(e), e) for e in l))

    def mset(self, *args):
        for i in range(0, len(args), 2):
            key = args[i]
            value = args[i + 1]
            self.dictionary[key] = value
        return b"+OK\r\n"


def main() -> int:
    print("Hello, World!")

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    loop = asyncio.get_event_loop()
    # Each client connection will create a new protocol instance
    coro = loop.create_server(RedisProtocol, "127.0.0.1", 7878)
    server = loop.run_until_complete(coro)

    # Serve requests until Ctrl+C is pressed
    print('Serving on {}'.format(server.sockets[0].getsockname()))
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass

    # Close the server
    server.close()
    loop.run_until_complete(server.wait_closed())
    loop.close()

    return 0


if __name__ == "__main__":
    sys.exit(main())
